/** MagicClass.  (JSX0.8)
	* Invoking the private native method allocationNewObject().

	**/

package JSX;
import java.lang.reflect.*;	//for Method etc
import java.io.*;	//for Serializable interface

class MagicClass {

	//Tuning: minimize object creation
		private final static Object[] argValues = new Object[2];
		private final static Class[] argTypes = {Class.class, Class.class};
		private final static Method allocateNewObject;	//in static, so can catch
		static {
			Method m = null;	//in static, so can catch
			try {
				m = java.io.ObjectInputStream.class.getDeclaredMethod(
					"allocateNewObject", argTypes);
				m.setAccessible(true); //the magic paint
			} catch (Exception e) {
				System.err.println(e);
				System.exit(0);
			}	
			allocateNewObject = m;	//ah!
		}


	/**------------------------------------------------------------------------
		* newInstance
		* -----------
		* Consider: subclass Class.
		**/
	/* We reuse the variable names from ObjectInputStream.inputObject(),
	 * purely for ease of reference with that method.
	 */
	static Object newInstance(Class currentClass) throws Exception, InvocationTargetException {

		//currentClass: is the actual runtime class we want to create.
		//currclass: is the closest non-serializable ancestor.
		// NB: Object is non-serializable, always bottoms out.
	//ONLY WORKS IF YOU ARE USING SERIALIZABLE CLASSES, I THINK!!!!
/*
		Class currclass = currentClass; 	//begin search for ancestor
		if (currclass != null) {
			while (Serializable.class.isAssignableFrom(currclass)) 
				currclass = currclass.getSuperclass();
		}		    
*/
		Class currclass = Object.class;	//always object
			//odd: why not use instanceOf?
 		argValues[0] = currentClass;	//faster than repeated creation of array
 		argValues[1] = currclass;
			//System.err.println(argValues[0]);
			//System.err.println(argValues[1]);

			//	allocateNewObject(currentClass, currclass);	//runtime link errror
		return allocateNewObject.invoke(null, argValues);
	}
	//could this possibly work?  Does the native code know where it is being
	//called from?
		//	static private native Object allocateNewObject(Class currentClass, Class currclass);
}

/**
	* All fields of serializable ancestor classes are filled in from stream,
	* so don't need a constructor to initialize them.

	* Note that if the most recent non-serializable ancestor lacks a public
	* no-arg constructor, it cannot be created.  It might have been better to
	* keep looking for one that works.

	* From the Java Documentation (this always confused me before):
	* Serialization does not read or assign values to the fields of any object
	* that does not implement the java.io.Serializable interface. Subclasses of
	* Objects that are not serializable can be serializable. In this case the
	* non-serializable class must have a no-arg constructor to allow its fields
	* to be initialized. In this case it is the responsibility of the subclass to
	* save and restore the state of the non-serializable class. It is frequently
	* the case that the fields of that class are accessible (public, package, or
	* protected) or that there are get and set methods that can be used to
	* restore the state. 
	**/
