package momoko;

import momoko.stream.*;
import momoko.protoclass.*;
import momoko.relationship.*;
import momoko.proxy.*;
import momoko.vfs.*;
import momoko.user.*;
import java.rmi.*;
import java.rmi.server.*;
import java.util.*;
import java.io.*;
import momoko.vfs.*;
import java.lang.reflect.*;
import java.beans.*;
import momoko.versioncontrol.*;

/**
 *  a multi-user aware object
 */
public class MObject implements IMObject, Serializable, Remote
{
  transient public Map relationships=new Hashtable();
  public Map subdir=new Hashtable();
  public IMObject mountpoint=null;
  public String mountpath=new String();
  public String namestr;
  public IMObject parent;
  static public boolean debug=false;

/**
 *  create a new object without a name
 */
public MObject()
{
  this("");
}

/**
 *  create a new object with the given name
 */
public MObject(String newname)
{
  namestr=newname;
}

public boolean release(IMObject object)  
{
  return true;
}

public boolean accept(IMObject object)  
{
  return true;
}

public void enterfunc(IMObject object) {}
public void exitfunc(IMObject object) {}

public VirtualFile getFileView()
{
  VirtualFile file=new ObjectFile(this);
//  if(debug) System.out.println("getFileView(): "+file);
  return file;
}

public VirtualFile getDirectoryView()
{
  VirtualFile file=new ObjectDirectory(this);
  return file;
}

public IMObject createLink(String oldPath) throws BadPathException, ClassException
{
  IMObject obj=Database.main.root.resolve(oldPath);
  obj=obj.makeLink();
  Database.main.move(obj, this);
  return obj;
}

public IMObject create(String ppath, String path, String name) throws IllegalArgumentException, BadPathException
{
  VersionController ver=(VersionController)Database.main.root.resolve("class/"+ppath);
  IProtoClass parent=ver.getCurrent();
  return create(parent, path, name);
}

public IMObject create(IProtoClass parent, String path, String name) throws IllegalArgumentException, BadPathException
{
if(debug) System.out.println("Create!");
  String prefix, suffix, temp=null;
  int i;
  boolean done=false;
  Iterator iterator;
  if(parent==null)
  {
    if(debug) System.out.println("Null parent!");
    return null;
  }
if(debug) System.out.println("create("+parent.getName()+", "+path+", "+name+")"); 
  if(debug) System.out.println("("+path+" "+name);
  if(path.equals(""))
  {
    if(debug) System.out.println(this.getName()+" will actually create the object.");
    return Database.main.addobj(parent, this, name);
  }
  i=path.indexOf('/');
  if(i==-1)
  {
    prefix=path;
    suffix="";
  }
  else if(i==0)
  {
    return Database.main.root.create(parent, path.substring(1), name);
  }
  else
  {
    prefix = path.substring(0, i);
    suffix = path.substring(i+1);
  }

  iterator = subdir.keySet().iterator();
  while(iterator.hasNext()&&!done)
  {
    temp=(String)iterator.next();
    if(debug) System.out.println("comparing "+prefix+" to "+temp);
    if(prefix.equals(temp))
      done=true;
  }
  if(done==false) throw new BadPathException(path);
  IMObject obj=(IMObject)subdir.get(temp);

  if(debug) System.out.println(this.getName()+" is passing on the torch to "+obj.getName());
  return obj.create(parent, suffix, name);
}

public IMObject makeLink() throws BadPathException
{
  return doMakeLink(this);
}

static public IMObject doMakeLink(IMObject o) throws BadPathException
{
  if(debug) System.out.println("makeLink");
  String path=o.getAbsolutePath();
  if(debug) System.out.println("path: "+path);
  String name=o.getName();
  return (IMObject)SoftProxy.newInstance(name, path);
}

public String getAbsolutePath()
{
  return doAbsolutePath(this);
}

static public String doAbsolutePath(IMObject o)
{
  if(debug) System.out.println("getAbsolutePath!!!!!!!!");
  if(debug) System.out.println("I am "+o);
  IMObject loc=o.location();
  if(debug) System.out.println("location: "+loc);
  if(loc==null) return "";
  else return loc.getAbsolutePath()+"/"+o.getName();
}

public IMObject resolve(String path) throws BadPathException
{
if(debug) System.out.println("resolving: "+path);
  String prefix, suffix, temp=null;
  int i;
  boolean done=false;
  Iterator iterator;
 
  i=path.indexOf('/');
  if(i==-1)
  {
    prefix=path;
    suffix="";
  }
  else if(i==0)
  {
    return Database.main.root.resolve(path.substring(1));
  }
  else
  {
    prefix = path.substring(0, i);
    suffix = path.substring(i+1);
  }
  iterator = subdir.keySet().iterator();
  while(iterator.hasNext()&&!done)
  {
    temp=(String)iterator.next();
    if(debug) System.out.println("comparing "+prefix+" to "+temp);
    if(prefix.equals(temp))
      done=true;
  }
  if(done==false) throw new BadPathException(this.getName()+"/"+path);
  Object o=subdir.get(temp);
  IMObject obj=(IMObject)o;

  if(suffix.equals(""))
    return obj;
  else
    return obj.resolve(suffix);
}

public void remove(String str)
{
  subdir.remove(str);
}

public void delete()
{
  System.out.println("delete() called on "+getName());
  System.out.println("location: "+location());
  Database.main.delete(this);
}

public String path()  
{
  String pathstr="";
  IMObject loc=this;
  while(!loc.equals(Database.main.root))
  {
    pathstr=loc.name()+pathstr;
    loc=loc.location();
  }
  pathstr="/"+pathstr;
  return pathstr;
}

/**
 *  is this object on this server or is it a proxy to a remote server?
 */
public boolean isLocal()
{
  return false;
}

public int compareTo(Object o)
{
  if(this.equals(o)) return 0;
  if(o instanceof MObject)
  {
    MObject mo=(MObject)o;
    String n1=this.getName();
    String n2=mo.getName();
    if(!n1.equals(n2))
      return n1.compareTo(n2);
  }
  Integer i1 = new Integer(this.hashCode());
  Integer i2 = new Integer(o.hashCode());
  return i1.compareTo(i2);
}

/**
 *  returns the object's relationships with other objects
 */
public Map getrelationships()
{
  return relationships;
}

public void setparent(IMObject p)
{
  parent=p;
}

/**
 *  returns the class the object is an instance of
 */
public IMObject getparent()
{
  return parent;
}

/**
 *  returns the object that contains this object
 */
public IMObject location()
{
  if(debug) System.out.println(this+" -- "+this.getName());
  return (IMObject)Database.main.contains.subject(this);
}

/**
 *  changes which object contains this object
 */
public void setlocation(IMObject newlocation)  
{
  Database.main.contains.relateforced(newlocation, this);
}

public Map getContents()
{
  return contents();
}

/**
 *  returns the subdirectory under this object in the directory tree
 */
public Map contents()
{
  if(Database.client) return subdir;
  Iterator iterator=subdir.keySet().iterator();
  while(iterator.hasNext())
  {
    String s=(String)iterator.next();
    Object o=subdir.get(s);
  }
  return subdir;
}

/**
 *  replaces the subdirectory under this object in the directory tree
 */
public void setcontents(Map newcontents)
{
  subdir=newcontents;
}

private void setmounts(IMObject obj, String s)  
{
  IMObject o;
  String s2;
  Iterator iterator;

  iterator=obj.contents().values().iterator();
  while(iterator.hasNext())
  {
    o=(IMObject)iterator.next();
    s2=s.concat(o.name());
    o.setmountpoint(this, s2);
    setmounts(o, s2);
  }
}

public void becomemount()  
{
  this.setmounts(this,"");
}

public void setmountpoint(IMObject point, String s)
{
  mountpoint=point;
  mountpath=s;
}

/**
 *  returns the object's name
 */
public String name()
{
  return namestr;
}

public String getName()
{
  return namestr;
}

public void setName(String n)
{
  namestr=n;
}

public void put(IMObject obj)
{
  subdir.put(obj.getName(), obj);
}

public void persistSelf()
{
  Map map=contents();
  synchronized(map)
  {
    Iterator iterator=map.values().iterator();
    while(iterator.hasNext())
    {
      IMObject obj=(IMObject)iterator.next();
      obj.persistSelf();
    }
  }
}

/**
 *  display object's relationships to other objects
 */
public void showself(StreamHandler handler)
{
  Iterator iterator;
  Enumeration iterator2;
  Object o;
  String p;
  java.util.Vector oarr;
  MObject mo;

  handler.println("relationships: "+relationships);
  if(relationships!=null)
  {
    iterator=relationships.keySet().iterator();
    while(iterator.hasNext())
    {
      p=(String)iterator.next();

      oarr=((Relationship)relationships.get(p)).objects(this);
      if(oarr!=null)
      {
        handler.print(p+" ->");
        iterator2=oarr.elements();
        while(iterator2.hasMoreElements())
        {
          mo=(MObject)iterator2.nextElement();
          handler.print(" "+mo.name());
        }
        handler.println("");
      }

      oarr=((Relationship)relationships.get(p)).subjects(this);
      if(oarr!=null)
      {
        iterator2=oarr.elements();
        handler.print(p+" <-");
        while(iterator2.hasMoreElements())
        {
          mo=(MObject)iterator2.nextElement();
          handler.print(" "+mo.name());
        }
        handler.println("");
      }
    }
  }
  else
    handler.println("no relationships");
  Class c = getClass();
  try
  {
  BeanInfo binfo = Introspector.getBeanInfo(c);
  PropertyDescriptor[] pd=binfo.getPropertyDescriptors();
  for(int x=0;x<pd.length;x++)
  {
    Method reader = pd[x].getReadMethod();
    if(reader==null)
    {
      handler.println("No reader for "+pd[x].getName());
      continue;
    }
    Object value=reader.invoke(this, null);
    handler.println(pd[x].getName()+" = "+value);
  }
  }
  catch(Exception e) {e.printStackTrace();}

/*  
  Field[] fields = c.getFields();
  for(int x=0;x<fields.length;x++)
  {
    Field f=fields[x];
    if(!Modifier.isStatic(f.getModifiers()))
      try { if(debug) System.out.println(f.getName()+" = "+f.get(this)); }
      catch(Exception e) {e.printStackTrace();}
  }
*/
}

}
