package momoko.extra;

import momoko.*;
import momoko.server.*;
import momoko.stream.*;
import momoko.user.*;
import java.util.*;
import momoko.forum.*;
import momoko.util.*;

public class PsuedoMOOInterpreter extends CommandInterpreter
{
  public PsuedoMOOInterpreter()
  {
    super();
    add("say", new say());
    add("emote", new emote());
    add("whisper", new whisper());
    add("look", new look());
    add("@dig", new dig());
    add("go", new go());
    add("@describe", new describe());
    add("@mail", new mail());
    add("@read", new read());
  }

  public void initialize()
  {
    handler.setparser(new LineParser());
  }

  public boolean interpret(Object command)
  {
    try { ((momoko.forum.User)avatar).setLastActionTime(new Date()); }
    catch(Exception e) {}
    try
    {
      String args=(String)command;
System.err.println(">>> "+command);
      String cmd=null;
      if(args.indexOf(' ')==-1) cmd=args;
      else cmd=(String)StringUtils.split(args, " ").elementAt(0);
      Object o=h.get(cmd);
      if(o==null) { handler.format("Unknown command: "+cmd); return true; }
      if(o instanceof MOOCommand)
      {
        MOOArguments margs=MOOParser.doparse(args);
        MOOCommand mc=(MOOCommand)o;
        mc.execute(margs, this);
      }
      else
      {
        String[] sargs=WordParser.doparse(args);
        Command c=(Command)o;
        c.execute(sargs, this);
      }
    }
    catch(ClassCastException e) {e.printStackTrace(); return false;}
    return true;
  }

  private class say implements MOOCommand
  {
    public void execute(MOOArguments args, Interpreter interp)
    {
    String s=args.subject;
    if (args.preposition.length() > 0) {
        s = s + " " + args.preposition;
    }
    if (args.object.length() > 0) {
        s = s + " " + args.object;
    }
    IMObject mo=null;
    mo=interp.getavatar().location();
    Room r=null;
    try
    {
      r=(Room)mo;
    }
    catch(ClassCastException e) {interp.handler.format("You are not in a room."); return;}
    r.say(interp.getavatar(), s);
    }
  }

  private class emote implements MOOCommand
  {
    public void execute(MOOArguments args, Interpreter interp)
    {
    String s=args.subject;
    if (args.preposition.length() > 0) {
        s = s + " " + args.preposition;
    }
    if (args.object.length() > 0) {
        s = s + " " + args.object;
    }
    IMObject mo=null;
    mo=interp.getavatar().location();
    Room r=null;
    try
    {
      r=(Room)mo;
    }
    catch(ClassCastException e) {interp.handler.format("You are not in a room."); return;}
    r.emote(interp.getavatar(), s);
    }
  }

  private class whisper implements MOOCommand
  {
    public void execute(MOOArguments args, Interpreter interp)
    {
    String s=args.preposition;
    if (args.object.length() > 0) {
        s = s + " " + args.object;
    }
    IMObject mo=null;
    mo=interp.getavatar().location();
    mo.contents(); // refresh user list
    IMObject oo=null;
    try
    {
      oo=mo.resolve(args.subject);
    }
    catch(BadPathException e)
    {
      handler.format("There is no one here by that name.");
      return;
    }

System.out.println("oo: "+oo);

    if(oo instanceof IUser)
    {
      IUser user=(IUser)oo;
System.out.println("invoking user.whisper");
      user.whisper(interp.getavatar(), s);
      handler.format("You whisper to "+oo.getName());
    }
    else
      handler.format(oo.getName()+" appears to be deaf.");
    }
  }

  private class mail implements MOOCommand
  {
    public void execute(MOOArguments args, Interpreter interp)
    {
try{
      String name=args.subject;
      MailFolder board=(MailFolder)Database.main.root.resolve("world/mail/mh/"+name);
      IFolder folder=board.getFolder();
      folder.open(javax.mail.Folder.READ_ONLY);
      IMessage[] messages = folder.getMessages();
      for(int x=0;x<messages.length;x++)
      {
        handler.format(new Integer(x+1).toString()+": "+messages[x].getSender()+" -- "+messages[x].getSubject());
      }
}
catch(Exception e) {e.printStackTrace();}
    }
  }

  private class read implements MOOCommand
  {
    public void execute(MOOArguments args, Interpreter interp)
    {
try{
      String num=args.subject;
      String name=args.object;
      IMObject board=Database.main.root.resolve("world/boards/"+name);
      IMessage message=(IMessage)board.resolve(num);
      handler.format("Subject: "+message.getSubject());
      handler.format("Sender:  "+message.getSender());
      handler.format("");
      handler.format(message.getContent());
      handler.format("");
}
catch(Exception e) {e.printStackTrace();}
    }
  }

  private class look implements MOOCommand
  {
    public void execute(MOOArguments args, Interpreter interp)
    {
    IMObject mo=null;
    IMObject avatar=null;
    try { avatar=interp.getavatar(); mo=avatar.location(); }
    catch(Exception e) {e.printStackTrace(); return;}
    Descriptive r=null;
    try
    {
      r=(Descriptive)mo;
      handler.format(r.getTitle());
      handler.format(r.getDescription());
    }
    catch(ClassCastException e)
    {
      System.out.println("System.out: "+System.out);
      System.out.println("Not descriptive...");
    }
    String items="";
      Iterator iterator=mo.contents().keySet().iterator();
      String o=avatar.getName();
      while(o==avatar.getName() && iterator.hasNext())
        o=(String)iterator.next();
      if(o==avatar.getName())
        return;
      else
        items="You see "+o;
      while(iterator.hasNext())
      {
        o=(String)iterator.next();
        if(o==avatar.getName()) continue;
        if(iterator.hasNext())
         items=items+", "+o;
        else
         items=items+" and "+o;
      }
    if(items!="") items=items+" here.";
    handler.format(items);
    }
  }

  private class go implements MOOCommand
  {
    public void execute(MOOArguments args, Interpreter interp)
    {
    IMObject mo=null;
    IMObject avatar=null;
    try { avatar=interp.getavatar(); mo=avatar.location(); }
    catch(Exception e) {e.printStackTrace(); return;}
    Navigable r=null;
    Map h=null;
    try
    {
      r=(Navigable)mo;
    }
    catch(ClassCastException e) {System.out.println("Location is not navigable."); return;}
    IMObject dest=r.getExit(args.subject);
    if(dest==null)
    {
      System.out.println("No such exit.");
      return;
    }
    try { Database.main.move(avatar, dest); }
    catch(Exception e) {handler.format("Move failed."); e.printStackTrace(); }
  }
 }

  private class dig implements MOOCommand
  {
    public void execute(MOOArguments args, Interpreter interp)
    {
    IMObject mo=null;
    IMObject avatar=null;
    try { avatar=interp.getavatar(); mo=avatar.location(); }
    catch(Exception e) {e.printStackTrace(); return;}
    Navigable r=null;
    try
    {
      r=(Navigable)mo;
    }
    catch(ClassCastException e) {System.out.println("Location is not navigable."); return;}
    IMObject dest=null;
    try { dest=mo.resolve(args.object); }
    catch(Exception e) { handler.format("Destination unreachable."); e.printStackTrace(); return; }
    if(dest==null)
    {
      System.out.println("Destination does not exist.");
      return;
    }
    r.setExit(args.subject, dest);
  }
 }

  private class describe implements MOOCommand
  {
    public void execute(MOOArguments args, Interpreter interp)
    {
    IMObject mo=null;
    IMObject avatar=null;
    try { avatar=interp.getavatar(); mo=avatar.location(); }
    catch(Exception e) {e.printStackTrace(); return;}
    Descriptive r=null;
    Map h=null;
    try
    {
      r=(Descriptive)mo;
    }
    catch(ClassCastException e) {System.out.println("Location is not describable."); return;}
    r.setDescription(args.object);
  }
 }
}
