/* -*- indent-tabs-mode: nil; c-basic-offset: 2 -*- */
package momoko.forum;

import momoko.util.*;
import java.rmi.*;
import java.rmi.server.*;
import java.util.*;
import java.io.*;
import momoko.*;
import java.lang.reflect.*;
import javax.mail.*;
import momoko.proxy.*;
import momoko.protoclass.*;
import momoko.forum.*;

/**
 *  a multi-user aware object
 */
public class GenericBoard extends GenericBoardInfo implements Board
{
  public static boolean debug=false;
  public static boolean speed_debug=true;
  transient boolean built=false;
  String path="/world/mail/maildir/";
  String infoPath=null;
  BoardInfo info=null;
  int messagecount = 0;
  Hashtable newness = new Hashtable();

/**
 *  create a new object without a name
 */
public GenericBoard()
{
  this("");
}

/**
 *  create a new object with the given name
 */
public GenericBoard(String newname)
{
  super(newname);
  setcontents(new OrderedMap());
}

private IMObject regenerate(String name)
{
  if(debug) System.out.println("GenericBoard.java - registering message "+name);
  IMObject server=null;
  try
  {
    if(subdir.size()==0)
      this.contents();
    server=(IMObject)subdir.get(name);
  }
  catch(Exception e) {e.printStackTrace(); return null;}
  return server;
}

public Map contents()
{
  int mc;
  IFolder board;

  try {
    MailFolder folder=
      (MailFolder)Database.main.root.resolve(path+getName());

    if(debug) System.out.println("Fetched mail folder: "+folder);

    if (speed_debug) System.out.println("Got folder -- "
                                        + new Date().toString());

    board=folder.getFolder();

    if (speed_debug) System.out.println("Got board -- "
                                        + new Date().toString());

    board.open(Folder.READ_WRITE);
    if(debug) System.out.println("Board.java - board: "+board);

    mc = board.getMessageCount();

    if(speed_debug) System.out.println(mc + " messages");

  } catch (Exception e) {
    e.printStackTrace();
    return super.contents();
  }


  // FIXME -- we only rethread if the message count has changed.
  // What we actually want to do is rethread if there have been any changes
  // to the folder at all.  Message count might not reflect that when
  // an equal number of adds and deletes have just happened, so we should
  // use a better metric.
  if ((!built) || (mc != this.messagecount)) {
    setcontents(new OrderedMap());
    buildThreads(board);
    this.messagecount = mc;
    built=true;
  }
  return super.contents();
}

public IMObject resolve(String path) throws BadPathException
{
  String prefix, suffix, temp=null;
  int i;
  boolean done=false;
  Enumeration iterator;

  // rethread if necessary
  contents();

  i=path.indexOf('/');
  if(i==-1)
  {
    prefix=path;
    suffix="";
  }
  else
  {
    prefix = path.substring(0, i);
    suffix = path.substring(i+1);
  }

  if(debug) System.out.println("Board.java - regenerating "+prefix);
  if(debug) System.out.println("Board.java - suffix: "+suffix);
  IMObject server=regenerate(prefix);
  if(debug) System.out.println("Board.java - result: "+server);
  if(server==null) throw new BadPathException(path);

  if(suffix.equals(""))
  {
    if(debug) System.out.println("Board.java - returning "+server);
    return server;
  }
  else
    return server.resolve(suffix);
}

public void persistSelf()
{
  //I'll decline, thanks.
}

public void setPath(String newPath)
{
  if(newPath==null) return;
  if(newPath.equals(path)) return;
  if(info!=null) info.setPath(newPath);
  path=newPath;
  setcontents(new OrderedMap());
  built=false;
}

public String getPath()
{
  return path;
}

public void setInfoPath(String newInfoPath)
{
  infoPath=newInfoPath;
  try
  {
    info=(BoardInfo)Database.main.root.resolve(newInfoPath);
    setPath(info.getPath());
  }
  catch(BadPathException e)
  {
    e.printStackTrace();
  }
}

public String getInfoPath() {return infoPath;}

public void setDescription(String s)
{
  if(info!=null) info.setDescription(s);
}

public String getDescription()
{
  if(info==null)
    return ""; 
  else 
    return info.getDescription();
}

public void setTitle(String s) {if(info!=null) info.setTitle(s);}
public String getTitle() {if(info==null) return ""; else return info.getTitle();}

/**
 * this is an implementation of the message threading algorithm
 * described at http://www.jwz.org/doc/threading.html
 */
public void buildThreads(IFolder board)
{
  if (speed_debug) System.out.println("Starting threading -- "
                                      + new Date().toString());
  int fakecount = 0;

  try
  {
    IMessage[] messages=board.getMessages();

    if (speed_debug) System.out.println("Starting threading -- got messages -- "
                                        + new Date().toString());

    if(debug) System.out.println("Board.java - messages: "+messages);
    if(debug) System.out.println("Board.java - count: "+messages.length);

    // id_table associates Message-IDs with Containers
    Hashtable id_table = new Hashtable(messages.length);

    if (speed_debug) System.out.println("Starting threading step 1 -- "
                                        + new Date().toString());
    // 1. For each message:
    for(int x=0; x<messages.length; x++)
    {
      ThreadContainer m = getContainer(messages[x], id_table);

      linkReferences(m, id_table, fakecount);
    }

    // FIXME -- this should happen due to return from a function call,
    // shouldn't it, Mister I-love-GC?
    messages = null;

    if(debug) System.out.println("Step 1 complete");

    if (speed_debug) System.out.println("Starting threading step 2 -- "
                                        + new Date().toString());
    // 2. Find the root set.  Walk over the elements of id_table, and
    // gather a list of the Container objects that have no parents.
    Vector root_set = findRootSet(id_table);

    if (speed_debug) System.out.println("Starting threading steps 3&4 -- "
                                        + new Date().toString());
    // 3. Discard id_table.  We don't need it anymore.
    // 4. Prune empty containers. Recursively walk all containers
    // under the root set.
    // (This doesn't need to be recursive, just iterate through all
    // the containers in id_table. --sah)
    root_set = pruneEmptyContainers(id_table, root_set);

    // *Now* we're done with id_table.  So discard it.  FIXME -- this
    // should happen due to return from a function call, shouldn't it,
    // Mister I-love-GC?
    id_table = null;

    if (speed_debug) System.out.println("Starting threading step 5 -- "
                                        + new Date().toString());
    // 5. Group root set by subject.
    groupBySubject(root_set, fakecount);

    // 6. Now you're done threading!

    // 7. Sort!
    if (speed_debug) System.out.println("Done threading: sorting -- "
                                        + new Date().toString());
    sortMessages(this, root_set, new MessageDateComparator());

    if(speed_debug) System.out.println("Completed threading -- "
                                        + new Date().toString());
  }
  catch(Exception e) {e.printStackTrace();}
  finally {built=true;}
}

private String getNameFromID(IMessage m)
{
  try { String[] temp=m.getHeader("Message-ID");
  String s=temp[0];
  Vector v=StringUtils.split(s, "-");
  return (String)v.elementAt(0);
  }
  catch(Exception e)
  {
    return new Integer(m.getMessageNumber()).toString();
  }
}

private ThreadContainer getContainer(IMessage message, Hashtable id_table)
  throws MessagingException
{
  String id = getId(message);
  ThreadContainer c=(ThreadContainer)id_table.get(id);
  ThreadContainer m=null;

  // A. If id_table contains a Container for this ID:
  if (c!=null)
  {
    // make an IMObject from the message
    IMObject mproxy=(IMObject)MObjectProxy.newInstance(
      new Integer(message.getMessageNumber()).toString(), message);
    Annotated annotation=(Annotated)newness.get(id);
    if(annotation==null)
    {
      annotation=new Annotation("N");
      newness.put(id, annotation);
    }
    mproxy=(IMObject)AnnotationProxy.newInstance(mproxy, annotation);
    mproxy.setcontents(new OrderedMap());

    m = new ThreadContainer((IMessage)mproxy);

    // now replace c with m:
    m.replace(c);
    id_table.remove(id);
    id_table.put(id, m);

  } else { // Else:
    IMObject mproxy=(IMObject)MObjectProxy.newInstance(
      new Integer(message.getMessageNumber()).toString(), message);
    Annotated annotation=(Annotated)newness.get(id);
    if(annotation==null)
    {
      annotation=new Annotation("N");
      newness.put(id, annotation);
    }
    mproxy=(IMObject)AnnotationProxy.newInstance(mproxy, annotation);
    mproxy.setcontents(new OrderedMap());

    m = new ThreadContainer((IMessage)mproxy);

    id_table.put(id, m);
    if(debug) System.out.println("Created a container with id "+id);
  }

  if(debug) System.out.println("Step 1A of threading complete");

  return m;
}


private String getId(IMessage message) throws MessagingException
{
  String idheaders[];

  idheaders = message.getHeader("Message-ID");
  if ((idheaders != null) &&
      (idheaders.length > 0))
  {
    return idheaders[0];
  }
  else
  {
    if (debug) System.out.println("Board.java - DAMN IT, resorting to using a totally fake message id");
    return new Integer(message.getMessageNumber()).toString();
  }
}


private void linkReferences(ThreadContainer m, Hashtable id_table, int fakecount)
  throws MessagingException
{
  // B. For each element in the message's Reference's field:
  Vector refs=getMessageReferences(m.getMessage());
  ListIterator iterator=refs.listIterator();

  ThreadContainer prev=null;
  while (iterator.hasNext())
  {
    String ref = (String)iterator.next();
    // Find a Container for the given Message-ID:
    // If there's one in id_table, use that
    ThreadContainer c=(ThreadContainer)id_table.get(ref);
    if (c==null)
    {
      // Otherwise, make (and index) one with a fake message
      String fakeId="fake-" + new Integer(fakecount).toString();
      fakecount++;
      IMObject message=FakeMessage.newInstance(fakeId);
      ((IMessage)message).setSubject("fake");
      c = new ThreadContainer((IMessage)message);
      id_table.put(ref, c);
      if (debug) System.out.println("Created placeholder message "+fakeId);
    }


    //  Link the References field's Containers together in the
    //  order implied by the References header.
    if (prev!=null)
    {
      // If they are already linked, don't change the existing links.
      if (c.getParent() == null)
      {
        // Do not add a link if adding that link would introduce a
        // loop: that is, before asserting A->B, search down the
        // children of B to see if A is reachable.
        //
        // (Actually, just search up A's parents looking for
        // B. --sah)

        if (!c.checkForDescendant(prev))
        {
          // prev->c
          c.move(prev);
        }
      }
    }

    prev=c;
  }

  if(debug) System.out.println("Step 1B of threading completed");

  // C. Set the parent of this message to be the last element in
  // References. Note that this message may have a parent already:
  // this can happen because we saw this ID in a References field,
  // and presumed a parent based on the other entries in that
  // field. Now that we have the actual message, we can be more
  // definitive, so throw away the old parent and use this new
  // one.

  // Note that this could cause this message to now have no
  // parent, if it has no references field, but some message
  // referred to it as the non-first element of its
  // references. (Which would have been some kind of lie...)

  // Do not add a link if adding that link would introduce a
  // loop: that is, before asserting A->B, search down the
  // children of B to see if A is reachable.
  //
  // (Actually, just search up A's parents looking for
  // B. --sah)

  if (!m.checkForDescendant(prev))
  {
    // prev->m
    m.move(prev);
  }
}


private Vector pruneEmptyContainers(Hashtable id_table, Vector root_set)
{
  Iterator iterator=id_table.values().iterator();
  // For each container:
  while(iterator.hasNext())
  {
    ThreadContainer c=(ThreadContainer)iterator.next();

    if (c.getMessage() instanceof Fake)
    {
      if(c.getChild() == null)
      {
        // A. If it is an empty container with no children, nuke it.
        c.move(null); // remove it from its parent
        root_set.remove(c); // remove it from the root set
      }
      else
      {
        // B. If the Container has no Message, but does have children,
        // remove this container but promote its children.
        ThreadContainer parent=c.getParent();

        // Do not promote the children if doing so would promote them to the
        // root set -- unless there is only one child, in which case, do.
        if ((parent != null)
            || (c.getChild().getNext() == null))
        {
          ThreadContainer current = c.getChild();

          while(current != null)
          {
            current.move(parent);
            current = current.getNext();
          }

          c.move(null); // remove it from its parent
          root_set.remove(c); // remove it from the root set
        }
      }
    }
  }

  return root_set;
}

private Vector findRootSet(Hashtable id_table)
{
  // 2. Find the root set.  Walk over the elements of id_table, and
  // gather a list of the Container objects that have no parents.
  Vector root_set = new Vector();
  Iterator iterator=id_table.values().iterator();
  while(iterator.hasNext())
  {
    ThreadContainer c=(ThreadContainer)iterator.next();

    if (c.getParent() == null)
    {
      if(debug) System.out.println("Found thread root: "
                                   +((IMObject)c.getMessage()).getName());
      root_set.add(c);
    }
  }

  return root_set;
}

private void groupBySubject (Vector root_set, int fakecount)
  throws MessagingException
{
  // If any two members of the root set have the same subject, merge
  // them. This is so that messages which don't have References
  // headers at all still get threaded (to the extent possible, at
  // least.)

  // A. Construct a new hash table, subject_table, which associates
  // subject strings with Container objects.
  Hashtable subject_table = new Hashtable(root_set.size());

  // B. For each Container in the root set:
  ListIterator literator=root_set.listIterator();
  while(literator.hasNext())
  {
    String subject = "";
    ThreadContainer c=(ThreadContainer)literator.next();

    // Find the subject of that sub-tree:
    if (!(c.getMessage() instanceof Fake))
    {
      // If there is a message in the Container, the subject is the
      // subject of that message.
      subject = c.getMessage().getSubject();
    }
    else
    {
      // If there is no message in the Container, then the Container
      // will have at least one child Container, and that Container
      // will have a message. Use the subject of that message
      // instead.
      subject = c.getChild().getMessage().getSubject();
      if (debug) System.out.println(subject);
    }

    // Strip ``Re:'', ``RE:'', ``RE[5]:'', ``Re: Re[4]: Re:'' and so on.
    subject = StringUtils.stripSubject(subject);

    // If the subject is now "", give up on this
    if ((subject != "") && (subject != null))
    {
      // this is a convenient time to set the subject of a fake to be the
      // subject of the thread
      if (c.getMessage() instanceof Fake)
      {
        c.getMessage().setSubject(subject);
      }

      // Add this Container to the subject_table if:

      ThreadContainer m=(ThreadContainer)subject_table.get(subject);
      if (m==null)
      {
        // There is no container in the table with this subject
        subject_table.put(subject, c);
      }
      else
      {
        if ((c.getMessage() instanceof Fake)
            && (!(m.getMessage() instanceof Fake)))
        {
          // or, this one is an empty container and the old one is
          // not: the empty one is more interesting as a root, so put
          // it in the table instead
          subject_table.remove(subject);
          subject_table.put(subject, c);
        }
        else if ((!StringUtils.detectReply(c.getMessage().getSubject()))
                 && (StringUtils.detectReply(m.getMessage().getSubject())))
        {
          // The container in the table has a ``Re:'' version of
          // this subject, and this container has a non-``Re:''
          // version of this subject. The non-re version is the more
          // interesting of the two.
          subject_table.remove(subject);
          subject_table.put(subject, c);
        }
      }
    }
  }

  // Now the subject_table is populated with one entry for each
  // subject which occurs in the root set. Now iterate over the root
  // set, and gather together the difference.

  // For each Container in the root set:
  Enumeration enum=root_set.elements();
  while(enum.hasMoreElements())
  {
    String subject = "";
    ThreadContainer c=(ThreadContainer)enum.nextElement();

    // Find the subject of this Container (as above.)

    // Find the subject of that sub-tree.  FIXME -- this is
    // cut-and-paste from above.  Make a method.
    if (!(c.getMessage() instanceof Fake))
    {
      // If there is a message in the Container, the subject is the
      // subject of that message.
      subject = c.getMessage().getSubject();
    }
    else
    {
      // If there is no message in the Container, then the Container
      // will have at least one child Container, and that Container
      // will have a message. Use the subject of that message
      // instead.
      subject = c.getChild().getMessage().getSubject();
    }

    // Strip ``Re:'', ``RE:'', ``RE[5]:'', ``Re: Re[4]: Re:'' and so on.
    subject = StringUtils.stripSubject(subject);

    // If the subject is now "", give up on this
    if ((subject != "") && (subject != null))
    {
      // Look up the Container of that subject in the table.
      ThreadContainer m = (ThreadContainer)subject_table.get(subject);

      // If it is null, or if it is *this* container, continue.
      if ((!(m==null)) && (!m.equals(c)))
      {
        // Otherwise, we want to group together this Container and
        // the one in the table. There are a few possibilities:

        if ((m.getMessage() instanceof Fake)
            && (c.getMessage() instanceof Fake))
        {
          // If both are dummies, append one's children to the
          // other, and remove the now-empty container.

          // move c's children to m
          ThreadContainer current = c.getChild();
          while(current != null)
          {
            current.move(m);
            current = current.getNext();
          }

          c.move(null); // remove it from its parent
          root_set.remove(c); // remove it from the root set
        }
        else if (m.getMessage() instanceof Fake)
        {
          // If one container is a empty and the other is not, make
          // the non-empty one be a child of the empty, and a
          // sibling of the other ``real'' messages with the same
          // subject (the empty's children.)

          c.move(m);
          root_set.remove(c); // remove it from the root set
        }
        else if (c.getMessage() instanceof Fake)
        {
          m.move(c);
          root_set.remove(m); // remove it from the root set
        }
        else if ((StringUtils.detectReply(c.getMessage().getSubject()))
                 && (!StringUtils.detectReply(m.getMessage().getSubject())))
        {
          // If that container is a non-empty, and that message's
          // subject does not begin with ``Re:'', but this message's
          // subject does, then make this be a child of the other.

          // If that container is a non-empty, and that message's
          // subject begins with ``Re:'', but this message's subject
          // does not, then make that be a child of this one -- they
          // were misordered. (This happens somewhat implicitly,
          // since if there are two messages, one with Re: and one
          // without, the one without will be in the hash table,
          // regardless of the order in which they were seen.)

          c.move(m);
          root_set.remove(c); // remove it from the root set
        }
        else
        {
          // Otherwise, make a new empty container and make both
          // msgs be a child of it. This catches the
          // both-are-replies and neither-are-replies cases, and
          // makes them be siblings instead of asserting a
          // hierarchical relationship which might not be true.

          String fakeId="fake-" + new Integer(fakecount).toString();
          fakecount++;
          IMObject message=FakeMessage.newInstance(fakeId);

          // set the subject of the fake to be the subject of the thread
          ((IMessage)message).setSubject(subject);

          ThreadContainer d = new ThreadContainer((IMessage)message);

          root_set.add(d);
          c.move(d);
          root_set.remove(c);
          m.move(d);
          root_set.remove(m);

          // hm, it seems like we should replace the entry in the
          // subject table with the new fake, although jwz doesn't
          // say so... --sah
          subject_table.remove(subject);
          subject_table.put(subject, d);
        }
      }
    }
  }
}


// Recursively sort messages according to a comparator.  Some of the
// comparators rely on the first child of a fake for ordering
// information, so it's important that this sort be depth-first,
// so that the first child of the fake will be the first *sorted*
// child of the fake.
private void sortMessages(IMObject container, Vector list, Comparator comparator)
  throws ClassException
{
  try {
    if (debug) System.out.println("Now sorting under: "
                                  + " (" +container.getName() + ")");
  } catch (Exception e) {}

  ListIterator iterator=list.listIterator();
  while(iterator.hasNext())
  {
    ThreadContainer c=(ThreadContainer)iterator.next();
    sortMessages(c, comparator);
  }

  Collections.sort(list, comparator);

  iterator=list.listIterator();
  while(iterator.hasNext())
  {
    IMObject c=(IMObject)((ThreadContainer)iterator.next()).getMessage();
    try {
      if (debug) System.out.println("Adding " +((IMessage)c).getSubject()
                                    + " (" + c.getName() + ")");
    } catch (Exception e) {}
    Database.main.move(c, container);
  }
}

private void sortMessages(ThreadContainer container, Comparator comparator)
  throws ClassException
{
  try {
    if (debug) System.out.println("Now sorting under: "
                                         +((IMessage)(container.getMessage())).getSubject()
                                  + " (" +((IMObject)(container.getMessage())).getName() + ")");
  } catch (Exception e) {}

  Vector list = new Vector();
  ThreadContainer current = container.getChild();
  while(current != null)
  {
    sortMessages(current, comparator);
    list.add(current);
    current = current.getNext();
  }

  Collections.sort(list, comparator);

  ListIterator iterator=list.listIterator();
  while(iterator.hasNext())
  {
    IMObject c=(IMObject)((ThreadContainer)iterator.next()).getMessage();
    try {
      if (debug) System.out.println("Adding " +((IMessage)c).getSubject()
                                    + " (" + c.getName() + ")");
    } catch (Exception e) {}
    Database.main.move(c, (IMObject)(container.getMessage()));
  }

}


// Get a list of references from the properly-parsed contents of
// In-Reply-To and References.
private Vector getMessageReferences(IMessage m) throws MessagingException
{
  String[] headers;

  headers=m.getHeader("References");
  if ((headers!=null) && (headers.length>0))
  {

    // Sigh, not every MUA separates references with newlines, so the
    // array we just got is useless.  Tack it all together and return
    // every angle-bracketed substring.

    int openangle = 0;
    int closeangle = 0;
    Vector refs = new Vector(headers.length);
    String refstring = "";

    for (int i=0; i<headers.length; i++)
    {
      refstring = refstring + headers[i];
    }

    do
    {
      openangle=refstring.indexOf('<', closeangle);
      if (openangle != -1)
      {
        closeangle=refstring.indexOf('>', openangle);

        if (closeangle != -1)
        {
          try
          {
            refs.add(refstring.substring(openangle, closeangle+1));
          }
          catch (IndexOutOfBoundsException e)
          {
            e.printStackTrace();
          }
        }
      }
    } while ((openangle != -1) && (closeangle != -1));


    return refs;
  }
  else
  {
    // If we don't have References, just use the first angle-bracketed
    // string in In-Reply-To.
    headers=m.getHeader("In-Reply-To");
    if ((headers!=null) && (headers.length>0))
    {
      int openangle;
      int closeangle;
      Vector refs = new Vector(1);

      openangle=headers[0].indexOf('<');
      if (openangle != -1)
      {
        closeangle=headers[0].indexOf('>', openangle);

        if (closeangle != -1)
        {
          try
          {
            refs.add(headers[0].substring(openangle, closeangle+1));
            return refs;
          }
          catch (IndexOutOfBoundsException e)
          {
            e.printStackTrace();
          }
        }
      }
    }
  }
  return new Vector(0);
}

private class StringIntComparator implements Comparator
{
  public int compare(Object o1, Object o2)
  {
    try
    {
      String s1=(String)o1;
      String s2=(String)o2;
      Integer i1=new Integer(s1);
      Integer i2=new Integer(s2);
      return i1.compareTo(i2);
    }
    catch(Exception e) { return 0; }
  }
}

private class MessageNumberComparator implements Comparator
{
  public int compare(Object o1, Object o2)
  {
    try
    {
      IMessage m1=(IMessage)((ThreadContainer)o1).getMessage();
      IMessage m2=(IMessage)((ThreadContainer)o2).getMessage();

      // if either message is a fake, get the first item in its
      // contents, and use that instead.
      if (m1 instanceof Fake)
      {
        m1=((IMessage)((IMObject)m1).contents().values().toArray()[0]);
      }

      if (m2 instanceof Fake)
      {
        m2=((IMessage)((IMObject)m2).contents().values().toArray()[0]);
      }

      Integer i1=new Integer(m1.getMessageNumber());
      Integer i2=new Integer(m2.getMessageNumber());

      return i2.compareTo(i1); // deliberately reverse the ordering
    }
    catch(Exception e) { return 0; }
  }
}

private class MessageDateComparator implements Comparator
{
  public int compare(Object o1, Object o2)
  {
    try
    {
      IMessage m1=(IMessage)((ThreadContainer)o1).getMessage();
      IMessage m2=(IMessage)((ThreadContainer)o2).getMessage();

      // if either message is a fake, get the first item in its
      // contents, and use that instead.
      if (m1 instanceof Fake)
      {
        m1=((IMessage)((IMObject)m1).contents().values().toArray()[0]);
      }

      if (m2 instanceof Fake)
      {
        m2=((IMessage)((IMObject)m2).contents().values().toArray()[0]);
      }

      Date d1=m1.getSentDate();
      Date d2=m2.getSentDate();

      return d2.compareTo(d1); // deliberately reverse the ordering
    }
    catch(Exception e) { return 0; }
  }
}

private class MessageSenderComparator implements Comparator
{
  public int compare(Object o1, Object o2)
  {
    try
    {
      IMessage m1=(IMessage)((ThreadContainer)o1).getMessage();
      IMessage m2=(IMessage)((ThreadContainer)o2).getMessage();
      String s1;
      String s2;

      if (m1 instanceof Fake)
      {
        s1 = "";
      }
      else
      {
        s1 = m1.getSender().toString();
      }

      if (m2 instanceof Fake)
      {
        s2 = "";
      }
      else
      {
        s2 = m2.getSender().toString();
      }

      return s1.compareTo(s2);
    }
    catch(Exception e) { return 0; }
  }
}

private class MessageSubjectComparator implements Comparator
{
  public int compare(Object o1, Object o2)
  {
    try
    {
      IMessage m1=(IMessage)((ThreadContainer)o1).getMessage();
      IMessage m2=(IMessage)((ThreadContainer)o2).getMessage();
      String s1=m1.getSubject();
      String s2=m2.getSubject();
      if (s1 == null) s1 = "";
      if (s2 == null) s2 = "";

      return s1.compareTo(s2);
    }
    catch(Exception e) { return 0; }
  }
}

}
