package momoko.forum;

import momoko.*;

// a Container object, more or less as described in
// http://www.jwz.org/doc/threading.html
public class ThreadContainer
{
    private IMessage message = null;
    private ThreadContainer parent = null;
    private ThreadContainer child = null; // first child
    private ThreadContainer next = null; // next element in sibling list, or null

    public ThreadContainer(IMessage message)
    {
        this.message = message;
    }

    public IMessage getMessage()
    {
        return message;
    }

    public ThreadContainer getParent()
    {
        return parent;
    }

    protected void setParent(ThreadContainer container)
    {
        parent = container;
    }

    public ThreadContainer getChild()
    {
        return child;
    }

    protected void setChild(ThreadContainer container)
    {
        child = container;
    }

    public ThreadContainer getNext()
    {
        return next;
    }

    protected void setNext(ThreadContainer container)
    {
        next = container;
    }

    // Check to see if this object has "container" as a descendant.
    // Since parent/child relationships should be consistant, it
    // should be safe to just check to see if "container" has this as
    // an ancestor.  Note that this method considers objects to be
    // descendants of themselves.
    public boolean checkForDescendant(ThreadContainer container)
    {
        boolean found = false;
        ThreadContainer current = container;
        while  (current != null) {
            if (current == this) {
                found = true;
                break;
            }
            current = current.getParent();
        }

        return found;
    }

    public void move(ThreadContainer container)
    {
        if (parent == container) {
            return;
        }

        if (parent != null) {
            // remove this object from its sibling list
            ThreadContainer current = parent.getChild();
            ThreadContainer prev;

            if (current == this) {
                parent.setChild(next);
            } else {
                do {
                    prev = current;
                    current = current.getNext();
                    if (current == null) {
                        // FIXME -- this would be really bad
                        System.out.println("AIEE, ThreadContainers are broken!");
                    }
                } while (current != this);

                prev.setNext(next);
            }
        }

        next = null;
        parent = container;

        if (parent != null) {
            // link this object to its new sibling list
            ThreadContainer current = parent.getChild();
            ThreadContainer prev;

            if (current == null) {
                parent.setChild(this);
            } else {
                do {
                    prev = current;
                    current = current.getNext();
                } while (current != null);

                prev.setNext(this);
            }
        }
    }

    // take the place of the given container
    // note: this method does not preserve sorting
    public void replace (ThreadContainer container)
    {
        // grab link to the old container's children, and reparent them
        child = container.getChild();
        ThreadContainer current = child;
        while (current != null) {
            current.setParent(this);
            current = current.getNext();
        }

        // move to the old container's parent
        move(container.getParent());

        // unlink the old container
        container.move(null);
    }
}
