package momoko.protoclass;

import java.rmi.*;
import java.util.*;
import java.io.*;
import java.lang.reflect.*;
import momoko.*;
import momoko.vfs.*;
//import momoko.compiler.jenesis.*;
import momoko.compiler.JavaCompiler;
import momoko.compiler.Compiler;
import momoko.versioncontrol.*;
import momoko.util.*;
import momoko.stream.*;
import momoko.proxy.*;

/**
 *  a working representation of a future Java class
 */
public class ProtoClass extends MObject implements IProtoClass, Remote
{
  static public boolean debug=false;
  private Hashtable properties=new Hashtable();
  private Vector children=new Vector();
  public String language="Java";
  private VersionController controller;
  private String parent;
  transient private AdaptiveClassLoader loader=Database.loader;

/**
 * create a new class without a name
 */
public ProtoClass()
{
  this("");
}

public ProtoClass(IProtoClass pc)
{
  super(pc.getName());
  properties=new Hashtable(pc.getproperties());
  children=new Vector(pc.getchildren());
}

/**
 *  create a new class with the given name
 */
public ProtoClass(String str)
{
  super(str);
}

public VirtualFile getFileView()
{
  File f=new File("source");
//  JenesisSourceWriter out=new JenesisSourceWriter();
//  out.write(f, this);
  String name=getName();
  name=name.replace('.', '/');
  return new RealFile(f.getAbsolutePath()+"/"+name+".java");
}

public VersionController getController()
{
  return controller;
}

public void setController(VersionController vc)
{
  controller=vc;
}

public IMObject getparent()
{
  return (IMObject)Database.main.parent.subject(this);
}

public String getParentName()
{
  return parent;
}

public void flatten()  
{
//  if(debug) System.out.println("FLATTEN!!");
  IMObject p=(IMObject)getparent();
//  if(debug) System.out.println("p#:*"+p.name());
  Enumeration iterator=((IProtoClass)p).getproperties().keys();
  while(iterator.hasMoreElements())
  {
    String name=(String)iterator.nextElement();
//    if(debug) System.out.println("parent has property "+name);
    properties.put(name, (Property)((IProtoClass)p).getproperties().get(name));
  }
}

public String getlanguage()
{
  return language;
}

public void setlanguage(String newlanguage)
{
  language=newlanguage;
}

/**
 *  returns the properties defined on this class
 */
public Hashtable getproperties()
{
  return properties;
}

/**
 *  make this class inherit from the specified parent class
 */
public void initialize(String parentname)
{
  IProtoClass parent;

//  if(debug) System.out.println("parent is "+parentname);
  try
  {
    VersionController vc=(VersionController)Database.main.resolve(Database.main.classroot,parentname);
    parent=vc.getCurrent();
  }
  catch(BadPathException e)
  {
    return;
  }
//  if(debug) System.out.println("parent is "+parent);
  Database.main.parent.relateforced(parent, this);
//  if(debug) System.out.println("forced relation");
}

/**
 *  returns all classes that inherit from this class
 */
public Vector getchildren()
{
  return children;
}

/**
 *  change the value of a defined property
 */
public void setprop(String key, Object newval) throws PermissionException
{
  Property p;
  Class t;

  p=(Property)properties.get(key);
  if(p!=null)
  {
    t=Database.types.named(p.type);
    if(!t.isInstance(newval))
      throw new PermissionException();
    this.setpropforreal(key, p.type, newval);  
  }
}

private void setpropforreal(String key, String type, Object obj)
{
  if(Database.types.checktype(type, obj))
    properties.put(key, new Property(type, obj));
  else
    if(debug) System.out.println("Not of type "+type);
}  

/**
 *  remove a defined property
 */
public void delprop(String key) throws PropertyNameException,
PermissionException
{
  Property info=(Property)properties.get(key);

  if(info==null)
    throw new PropertyNameException();
  properties.remove(key);
}

/**
 *  define a new property
 */
public void addprop(String key, String type, Object val) throws
PropertyNameException, PermissionException
{
  Property p;

  p=(Property)properties.get(key);
  if(p!=null)
    throw new PropertyNameException();
  this.setpropforreal(key,type, val);
}

/**
 *  get the value of a property based on its name
 */
public Object getprop(String key) throws PropertyNameException
{
  Property p;
  Object o;

  p=(Property)properties.get(key);
  if(p!=null)
  {
    return p.object;
  }
  else
    throw new PropertyNameException();
}

/**
 *  display defined properties
 */
public void showself(StreamHandler handler)
{
  Enumeration iterator;
  Enumeration iterator2;
  Object o;
  Pair p;
  Property prop;
  Vector oarr;
  MObject mo;

  handler.println(this.name()+':');
  iterator=properties.keys();
  while(iterator.hasMoreElements())
  {
    o=iterator.nextElement();
    prop=(Property)properties.get(o);
    handler.println(prop.type+" "+o+"="+prop.object);
  }
  super.showself(handler);
}

public AdaptiveClassLoader getClassLoader()
{ 
  if(loader==null)
  {
    ClassRepository cr=(ClassRepository)controller.location();
    loader=cr.getClassLoader();
  }
  return loader; 
}
public void setClassLoader(AdaptiveClassLoader newLoader) {loader=newLoader;}

public Class getClassObject() throws ClassNotFoundException
{
  if(loader==null)
  {
    ClassRepository cr=(ClassRepository)controller.location();
    if(debug) System.out.println("cr: "+cr);
    loader=cr.getClassLoader();
  }
//  if(debug) System.out.println("$$$ "+loader+" "+getName());
  return loader.loadClass(getName());
}

public byte[] getClassBytes() throws ClassNotFoundException
{
  System.out.println("ProtoClass.getClassBytes()");
  if(loader==null)
  {
    ClassRepository cr=(ClassRepository)controller.location();
//if(debug) System.out.println("My location is "+cr);
    loader=cr.getClassLoader();
    System.out.println("loader: "+loader);
  }
  System.out.println("got loader: "+loader);
  byte[] b=loader.loadClassBytes(getName());
  System.out.println("b: "+b);
  return b;
}

public IMObject newInstance(String name)
{
  try{
  Class[] classes=new Class[1];
  Object[] objects=new Object[1];
  classes[0]=String.class;
  objects[0]=name;
  Constructor constructor=getClassObject().getConstructor(classes);
//if(debug) System.out.println("constructor "+constructor);
  Object o=constructor.newInstance(objects);
  if(o instanceof IMObject) return (IMObject)o;
  else
  {
    System.out.println("New object "+name+" is not an IMObject. Making it one by wrapping in a proxy.");
    return MObjectProxy.newInstance(name, o);
  }
  }catch(Exception e) {e.printStackTrace(); return null;}
}

}
