/*
 * Copyright (c) 1997-1998 The Java Apache Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Java Apache 
 *    Project for use in the Apache JServ servlet engine project
 *    (http://java.apache.org/)."
 *
 * 4. The names "Apache JServ", "Apache JServ Servlet Engine" and 
 *    "Java Apache Project" must not be used to endorse or promote products 
 *    derived from this software without prior written permission.
 *
 * 5. Products derived from this software may not be called "Apache JServ"
 *    nor may "Apache" nor "Apache JServ" appear in their names without 
 *    prior written permission of the Java Apache Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Java Apache 
 *    Project for use in the Apache JServ servlet engine project
 *    (http://java.apache.org/)."
 *    
 * THIS SOFTWARE IS PROVIDED BY THE JAVA APACHE PROJECT "AS IS" AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE JAVA APACHE PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Java Apache Group. For more information
 * on the Java Apache Project and the Apache JServ Servlet Engine project,
 * please see <http://java.apache.org/>.
 *
 */

package momoko.protoclass;

import momoko.*;
import momoko.util.*;
import momoko.webserver.*;
import momoko.versioncontrol.*;

import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;
import java.util.zip.*;

public class RemoteClassLoader extends AdaptiveClassLoader implements Serializable {

    static boolean debug=false;
    private IMObject rep;

    public RemoteClassLoader(IMObject classRepository, Vector v)
        throws IllegalArgumentException
    {
        super(v);

        // Verify that all the repository are valid.
        Iterator e = classRepository.contents().values().iterator();
        while(e.hasNext()) {
            Object o = e.next();
            VersionController file;

            // Check to see if element is a File instance.
            try {
                file = (VersionController) o;
            } catch (ClassCastException f) {
                throw new IllegalArgumentException("Object " + o
                    + "is not a valid \"VersionController\" instance");
            }
        }

        // Store the class repository for use
        this.rep = classRepository;

        // Increment and store generation counter
        this.generation = generationCounter++;
    }

    //------------------------------------------------------- Methods

    public Vector listContents()
    {
      Vector v=new Vector();
      Iterator iterator=rep.contents().values().iterator();
      while(iterator.hasNext())
      {
        VersionController file=(VersionController)iterator.next();
        v.addElement(file.getName());
      }
      return v;
    }

    public synchronized boolean shouldReload(String classname) {

        return true;

     }

    public synchronized boolean shouldReload() {

        return true;

    }

    public AdaptiveClassLoader reinstantiate() {
        return new RemoteClassLoader(rep, repository);
    }

    protected synchronized Class loadClass(String name, boolean resolve)
        throws ClassNotFoundException
    {
        // The class object that will be returned.
        Class c = null;

        // Use the cached value, if this class is already loaded into
        // this classloader.
        ClassCacheEntry entry = (ClassCacheEntry) cache.get(name);

        if (entry != null) {
            if(debug) System.out.println("Entry cached");
            // Class found in our cache
            c = entry.loadedClass;
            if (resolve) resolveClass(c);
            return c;
        }


        try
        { 
          if(debug) System.out.println("Attempting local load");
          Class local = super.loadClass(name, resolve);
          if(local!=null)
          {
            return local;
          }
        }
        catch(ClassNotFoundException e) {}

        ClassCacheEntry classCache = new ClassCacheEntry();

        byte[] classData=null;

        if(debug) System.out.println("Loading class byte remotely");

        classData = loadClassBytes(name);

            if (classData != null) {

                if(debug) System.out.println("Successfully remotely fetched the class.");
                // Define the class
                c = defineClass(name, classData, 0, classData.length);

                // Cache the result;
                classCache.loadedClass = c;
                cache.put(name, classCache);
    
                // Resolve it if necessary
                if (resolve) resolveClass(c);
                
                return c;
            }

            if(debug) System.out.println("Failed to remotely load the class.");

            throw new ClassNotFoundException(name);
    }

    public byte[] loadClassBytes(String name) throws ClassNotFoundException
    {
/*
        try { return super.loadClassBytes(name); }
        catch(Exception e) {}
*/

        try
        { 
          if(debug) System.out.println("Resolving VC");
          VersionController file = (VersionController)rep.resolve(name);
          if(debug) System.out.println("VC: "+file);
          IProtoClass current=file.getCurrent();
          if(debug) System.out.println("current: "+current);
          byte[] classData = current.getClassBytes();
          if(debug) System.out.println("classData: "+classData);
          if(classData==null) throw new ClassNotFoundException(name);
          else return classData;
        }
        catch(BadPathException e) {throw new ClassNotFoundException(name);}
    }

    public void defineClass(String name, byte[] classData)
    {
                ClassCacheEntry classCache = new ClassCacheEntry();
                Class c = defineClass(name, classData, 0, classData.length);
                // Cache the result;
                classCache.loadedClass = c;
                cache.put(name, classCache);
    }

    public URL getResource(String name)
    {
System.out.println("getResource("+name+")");
      URL url=super.getResource(name);
      if(url==null)
      {
        try
        {
          System.out.println("Fetching URL for remote class");
          Object temp=rep.resolve(name);
          System.out.println("resource: "+temp);
          PrintUtils.printObject(temp);
          Resource resource=(Resource)temp;
          System.out.println("resource: "+resource);
          url=resource.getURL();
        }
        catch(Exception e)
        {
          e.printStackTrace();
          return null;
        }
      }
      return url;
    }
}
