package momoko.proxy;

import momoko.*;
import momoko.util.*;
import java.lang.reflect.*;
import java.io.*;

public class DelegationProxy implements InvocationHandler, Serializable 
{
    static public boolean debug=false;

    // preloaded Method objects for the methods in java.lang.Object

    public Class[] interfaces;
    public Object[] delegates;

    public DelegationProxy() {this.interfaces=null; this.delegates=null;}

    public DelegationProxy(Class[] interfaces, Object[] delegates) {
        this.interfaces = (Class[]) interfaces.clone();
        this.delegates = (Object[]) delegates.clone();
    }

    public Object invoke(Object proxy, Method m, Object[] args)
        throws Throwable
    {
        Class declaringClass = m.getDeclaringClass();

        if(m.getName().equals("toString"))
          return proxyToString(proxy);


if(debug) System.out.print("invoke("+m+", [");
if(args!=null)
  for(int x=0;x<args.length;x++)
    if(debug) System.out.print(args[x]+",");
if(debug) System.out.println("])");

        if(m.getName().equals("writeObject")) 
{ writeObject((ObjectOutput)args[0]); return null; }
        if(m.getName().equals("readObject")) 
{ readObject((ObjectInput)args[0]); return null; }

if(m.getName().equals("hashCode")) return proxyHashCode(proxy);
if(m.getName().equals("equals")) return proxyEquals(proxy, args);
if(m.getName().equals("location")) return Database.main.contains.subject(proxy);
if(m.getName().equals("delete"))
{
  Database.main.delete((IMObject)proxy);
  return Void.TYPE;
}

            for (int i = 0; i < interfaces.length; i++) {
                if (declaringClass.isAssignableFrom(interfaces[i])) {
                    try {
                        if(debug) System.out.println("Delegating "+m+" to "+delegates[i]);
                        return m.invoke(delegates[i], args);
                    } catch (InvocationTargetException e) {
                        throw e.getTargetException();
                    }
                }
            }

            return invokeNotDelegated(proxy, m, args);
    }

    protected Object invokeNotDelegated(Object proxy, Method m,
                                        Object[] args)
        throws Throwable
    {
        throw new InternalError("unexpected method dispatched: " + m);
    }

    protected Integer proxyHashCode(Object proxy) {
if(debug) System.out.println(System.identityHashCode(proxy));
        return new Integer(System.identityHashCode(proxy));
    }

    protected Boolean proxyEquals(Object proxy, Object other)
    {
      if(debug) System.out.println("proxyEquals");
      if(debug) System.out.println("proxy: "+proxy);
      if(debug) System.out.println("other: "+other);
      if(debug) System.out.println(proxy==other);
      Class c=other.getClass();
      if(c.isArray())
      {
        Object[] oa = (Object[])other;
        if(debug) System.out.println("Array: "+oa.length);
        if(debug) System.out.println("{");
        for(int x=0;x<oa.length;x++)
          if(debug) System.out.println(oa[x]);
        if(debug) System.out.println("}");
        if(oa.length==1)                            // test kludge
          if(oa[0]==proxy) return Boolean.TRUE;      // test kludge
      }
      return (proxy == other ? Boolean.TRUE : Boolean.FALSE);
//      return Boolean.TRUE; // test kludge
    }

    protected String proxyToString(Object proxy) {
        String s="DelegationProxy: [";
        for(int x=0;x<delegates.length;x++)
          s=s+delegates[x]+",";
        s=s+"]";
        return s;
    }

  public void readObject(ObjectInput in)
  {
    if(debug) System.out.println("DelegationProxy.readObject()");
    try { Pair p=(Pair)in.readObject();
    interfaces=(Class[])p.first;
    delegates=(Object[])p.second;
    }catch(Exception e) {e.printStackTrace();}
  }

  public void writeObject(ObjectOutput out)
  {
    try { out.writeObject(new Pair(interfaces, delegates)); }
    catch(Exception e) {e.printStackTrace();}
  }
}
