package momoko.server;

import momoko.vfs.*;
import java.io.*;
import momoko.*;
import momoko.stream.*;
import momoko.user.*;
import momoko.protoclass.*;
import java.util.*;
import java.lang.reflect.*;
import momoko.forum.User;
import momoko.util.*;
import momoko.compiler.Compiler;
import momoko.proxy.*;
import momoko.forum.*;

public class CommandInterpreter extends Interpreter
{
  protected IAdmin avatar;
  protected String cwd;
  protected IProtoClass working;
  protected Function f;
  protected String fname;
  protected Vector arr;
  protected Compiler comp;
  protected Hashtable h=new Hashtable();
  protected boolean running=false;
  protected static boolean debug = true;

  public CommandInterpreter()
  {
    super();
    add("printdb", new printdb());
    add("@move", new move());
    add("who", new who());
    h.put("look", new look());
    h.put("show", new show());
    h.put("@create", new newobj());
    h.put("import", new importObj());
    h.put("class", new newclass());
    h.put("lang", new lang());
    h.put("add", new add());
    h.put("program", new program());
    h.put("return", new ret());
    h.put("arg", new arg());
    h.put(".", new dot());
    h.put("compile", new compile());
    add("@set", new set());
    add("@get", new get());
    h.put("#", new comment());
    h.put("shutdown", new shutdown());
    h.put("ln", new link());
    add("@quit", new quit());
    add("scheme", new scheme());
    add("@dump-database", new dumpdb());
    add("rm", new rm());
    add("showboard", new showboard());
    add("cat", new cat());
  }

  public void add(String name, Object cmd)
  {
    h.put(name, cmd);
  }

  public void setavatar(IUser av)
  {
    avatar=(IAdmin)av;
  }

  public IMObject getavatar()
  {
    return avatar;
  }

  public boolean interpret(Object command)
  {
    try { ((momoko.forum.User)avatar).setLastActionTime(new Date()); }
    catch(Exception e) {}
    try
    {
      String[] args=(String[])command;
if(debug) System.err.println(">-> "+args[0]);
      Command c=(Command)h.get(args[0]);
      if(c==null) handler.format("Unknown command: "+args[0]);
      else c.execute(args, this);
    }
    catch(ClassCastException e) {return false;}
if(debug) System.out.println("running: "+running);
    return running;
  }

public void close() {if(debug) System.out.println("closing");running=false;}

protected String getarg(String[] args, int i)
{
  String s="";
  try { s=args[i]; } catch(ArrayIndexOutOfBoundsException e) {}
  return s;
}

private class link implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
    try{
    String source=command[1];
    String dest=command[2];
    String name=command[3];
    IMObject dobj=Database.main.root.resolve(dest);
    IMObject link=(IMObject)SoftProxy.newInstance(name, source);
    Database.main.move(link, dobj);
    }catch(Exception e) {e.printStackTrace(); handler.format("Link failed.");}
  }
}

private class scheme implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
    handler.setinterpreter(new SchemeInterpreter());
  }
}

private class dumpdb implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
    Date d=new Date();
    String dstr=d.toString();
    int dlen=dstr.length();
    handler.println("*        Checkpoint started at "+d+"        *");
    Database.main.root.persistSelf();
    Date d2=new Date();
    long time=d2.getTime()-d.getTime();
    dlen=dlen-new Long(time).toString().length();
    String spacing="";
    for(int x=0;x<dlen;x++) spacing=spacing+" ";
    handler.println("*        Finished in "+time+"      "+spacing+"*");
  }
}

private class rm implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
    try
    {
      String path=command[1];
      String oname=command[2];
      IMObject o=Database.main.root.resolve(path);
      o.remove(oname);
    }
    catch(Exception e)
    {
      handler.format(e);
    }
  }
}

private class showboard implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
    try
    {
      String path=command[1];
      Board b=(Board)Database.main.root.resolve(path);
      Map map=b.contents();
      printMessages(map, "");
    }
    catch(Exception e)
    {
      handler.format(e);
    }
  }
  private void printMessages(Map map, String prefix) throws Exception
  {
    Iterator iterator=map.values().iterator();
    while(iterator.hasNext())
    {
      Object o=iterator.next();
      IMessage m=(IMessage)o;
      handler.format(prefix+m.getSubject());
      IMObject mobj=(IMObject)o;
      printMessages(mobj.contents(), prefix+" ");
    }
  }
}

private class cat implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
    try
    {
      String path=command[1];
      VirtualFile f=(VirtualFile)Database.main.root.resolve(path);
      InputStream in=f.getInputStream();
      new Conduit(in, handler.getOutputStream()).run();
    }
    catch(Exception e)
    {
      handler.format(e);
    }
  }
}

private class printdb implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
//    ((SubjectExclusiveRelationship)Database.main.contains).print();
  }
}

private class who implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
    try{
      IMObject db=Database.main.root.resolve("world/db");
      Map contents=db.contents();
      boolean found=false;
      Iterator iterator=contents.values().iterator();
      while(iterator.hasNext())
      {
        User user=(User)iterator.next();
        if(user.getConnectedStatus())
        {
          found=true;
          long idle = user.getIdleTime();
          idle=idle/1000;
          handler.format(user.getName()+", idle "+idle+" seconds");
        }
      }
      if(!found)
        handler.format("No one is here.");
    }catch(Exception e) {e.printStackTrace(); handler.format("Failed.");}
  }
}

private class comment implements Command
{
  public void execute(String[] command, CommandInterpreter interp) {}
}

private class move implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
    {
        if(debug) System.out.println("Attempting move");
        interp.avatar.move(getarg(command, 1), getarg(command, 2));
    }
}

private class help implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
    {
        interp.handler.format("\nObject manipulation:");
        interp.handler.format("look");
        interp.handler.format("show {obj}");
        interp.handler.format("new parent name");
        interp.handler.format("get path");
        interp.handler.format("\nClass manipulation:");
        interp.handler.format("class parent name");
        interp.handler.format("lang language");
        interp.handler.format("put name type {value}");
        interp.handler.format("\nProgramming:");
        interp.handler.format("program name");
        interp.handler.format("return type");
        interp.handler.format("arg type name");
        interp.handler.format(". code");
        interp.handler.format("compile");
        interp.handler.format("\nSystem:");
        interp.handler.format("shutdown");
        interp.handler.format("\nOther modes:");
        interp.handler.format("python");
    }
}

private class look implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
    {
        if(debug) System.out.println("looking");
        IMObject r=interp.avatar.location();
        if(debug) System.out.println(r);
    }
}

private class shutdown implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
    {
      MudServer.shutdown();
    }
}

private class show implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
      try
      {
        IMObject o=interp.avatar.getobj(command[1]);
        System.out.println("show o: "+o);
        if(o==null)
          handler.format("No such object");
        else
          o.showself(handler);
      }
      catch(ArrayIndexOutOfBoundsException e)
      {
        Database.main.tree(handler);
        handler.format(".");
      }
  }
}

private class newobj implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
      String cls=command[1];
      String loc=null;
      String name=null;
      try
      {
        loc=command[2];
        name=command[3];
      }
      catch(ArrayIndexOutOfBoundsException e)
      {
        loc=null;
        name=command[2];
      }
      interp.avatar.newobj(cls, loc, name);
      System.out.println("New object created ("+name+")");
  }
}

private class importObj implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
      String cls=command[1];
      String loc=null;
      String name=null;
      try
      {
        loc=command[2];
        name=command[3];
      }
      catch(ArrayIndexOutOfBoundsException e)
      {
        loc=null;
        name=command[2];
      }
      interp.avatar.importObj(cls, loc, name);
  }
}

private class newclass implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
      working=interp.avatar.newclass(getarg(command, 1),getarg(command,2));
  }
}

private class compiler implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
      try { comp=(Compiler)Database.main.resolve(Database.main.root, getarg(command,1)); } catch(Exception e) {e.printStackTrace();}
  }
}

private class compile implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
    {
        try { interp.avatar.compile(getarg(command,1), comp); } catch(Exception e) {}
    }
  }
}

private class newversion implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
      working=(IProtoClass)interp.avatar.newversion(getarg(command,1));
  }
}

private class lang implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
      if(working!=null) working.setlanguage(getarg(command,1));
  }
}

private class add implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
      if(debug) System.out.println("?");
      try{working.addprop(getarg(command,1),getarg(command,2),Database.types.convert(getarg(command,2),getarg(command,3)));}
      catch(Exception e) {if(debug) System.out.println(e);}
      if(debug) System.out.println("Perhaps I added.");
  }
}

private class quit implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
    interp.close();
  }
}

private class program implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
      try
      {
        fname=getarg(command,1);
        f=(Function)working.getprop(fname);
      }
      catch(PropertyNameException e) {}
      if(!Database.types.checktype("Function", f))
      {
        f=null;
        if(debug) System.out.println("Not a function");
      }
  }
}

private class ret implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
      if(f==null)
        if(debug) System.out.println("No function");
      else
      {
        f.set_returnType(getarg(command,1));
        try { working.setprop(fname, f); }
        catch(PermissionException e) {}
      }
  }
}

private class arg implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
      if(f==null)
        if(debug) System.out.println("No function");
      else
      {
        arr=f.get_args();
        arr.addElement(new Pair(getarg(command,1), getarg(command,2)));
        f.set_args(arr);
        try {working.setprop(fname, f);}
        catch(PermissionException e) {}
      }
  }
}

private class dot implements Command
{
  public void execute(String[] command, CommandInterpreter interp)
  {
      if(f==null)
        if(debug) System.out.println("No function");
      else if(working==null)
        if(debug) System.out.println("No object");
      else
      {
        if(debug) System.out.println("it is in f adding");
        arr=f.get_sourceCode();
        if(debug) System.out.println("but after source getting");
        arr.addElement(getarg(command,1)+" "+getarg(command,2)+" "+getarg(command,3));
        f.set_sourceCode(arr);
        try
        {
          if(debug) System.out.println("to set");
          working.setprop(fname, f);
        }
        catch(PermissionException e) {}
      }
 }
}

private class set implements Command
{
  private String intercap(String s)
  {
    String first=s.substring(0,1).toUpperCase();
    String second=s.substring(1,s.length());
    return first+second;
  }

  public void execute(String[] command, CommandInterpreter interp)
  {
try{
    String name=getarg(command,1);
    String prop=getarg(command,2);
    Object value=Database.types.convert(getarg(command,3),getarg(command,4));
    IMObject obj=Database.main.resolve(avatar.location(), name);
    System.out.println("set obj: "+obj);
    Class[] carr=new Class[1]; carr[0]=value.getClass();
    Object[] oarr=new Object[1]; oarr[0]=value;
    Method setter=obj.getClass().getMethod("set"+intercap(prop), carr);
    setter.invoke(obj, oarr);
}
catch(Exception e) {e.printStackTrace();}
 }
}

private class get implements Command
{
  private String intercap(String s)
  {
    String first=s.substring(0,1).toUpperCase();
    String second=s.substring(1,s.length());
    return first+second;
  }

  public void execute(String[] command, CommandInterpreter interp)
  {
try{
    String name=getarg(command,1);
    String prop=getarg(command,2);
    IMObject obj=Database.main.resolve(avatar.location(), name);
    System.out.println("get obj: "+obj);
    Class[] carr=new Class[0];
    Object[] oarr=new Object[0];
    Method getter=obj.getClass().getMethod("get"+intercap(prop), carr);
    handler.println(getter.invoke(obj, oarr));
}
catch(Exception e) {e.printStackTrace();}
 }
}

}
