package momoko.server;

import java.net.*;
import java.util.*;
import momoko.*;
import java.io.*;
import momoko.drivers.*;
import momoko.stream.*;

public class SocketService implements Service
{
  boolean running=true;
  ServerSocket listener;
  Vector sockets=new Vector();
  Vector threads=new Vector();

  private class SocketListener implements Runnable
  {
    Socket socket;
    Parser parser;
    Formatter formatter;
    Interpreter interpreter;
    StreamHandler handler;
    Thread thread;

    public void run()
    {
      while(running)
      {
        try
        {
          socket=listener.accept();
System.out.println("Accepted connection");
          sockets.addElement(socket);
          parser=(Parser)parserclass.newInstance();
          formatter=(Formatter)formatterclass.newInstance();
          interpreter=(Interpreter)interpreterclass.newInstance();
          handler=new StreamHandler();
	  handler.setparser(parser);
	  handler.setformatter(formatter);
	  handler.setinstream(socket.getInputStream());
          OutputStream os=socket.getOutputStream();
	  handler.setoutstream(os);
	  handler.setinterpreter(interpreter);
	  handler.initialize();
          thread=new RedirectingThread(os, new SocketReader(handler, socket), "socket reader");
          threads.addElement(thread);
          thread.start();
        }
        catch(Exception e)
        {
          if(MudServer.running)
          {
            System.out.println("Error starting service on new socket");
            e.printStackTrace();
          }
        }
      }
    }
  }

  private class SocketReader implements Runnable
  {
    StreamHandler handler;
    Socket socket;

    public SocketReader(StreamHandler sh, Socket s)
    {handler=sh; socket=s;}

    public void run()
    {
      System.out.println("Reading on a new socket");
      handler.handle();
      try { socket.close(); } catch(IOException e) {}
    }
  }

  String str;
  int port;
  Class parserclass;
  Class formatterclass;
  Class interpreterclass;
  Thread listenthread;

  public void initialize(Properties args) throws
  UnspecifiedArguementException, BadArguementException, IOException,
  DriverNotFoundException
  {
    str=args.getProperty("port");
    if(str==null)
      throw new UnspecifiedArguementException();
    try{port=new Integer(str).intValue();}
    catch(NumberFormatException e)
    {throw new BadArguementException();}
    str=args.getProperty("parser");
    if(str==null)
      throw new UnspecifiedArguementException();
    str="momoko.server."+str+"Parser";
    try {parserclass=Class.forName(str);}
    catch(Exception e) {throw new DriverNotFoundException(str);}
    str=args.getProperty("interpreter");
    if(str==null)
      throw new UnspecifiedArguementException();
    try {interpreterclass=Class.forName("momoko.server."+str+"Interpreter");}
    catch(Exception f)
    {
      try {interpreterclass=Class.forName(str);}
      catch(Exception e) {throw new DriverNotFoundException(str);}
    }
    str=args.getProperty("formatter");
    if(str==null)
      throw new UnspecifiedArguementException();
    str="momoko.server."+str+"Formatter";
    try {formatterclass=Class.forName(str);}
    catch(Exception e) {throw new DriverNotFoundException(str);}
    listener=new ServerSocket(port);
  }

  public void service()
  {
    listenthread=new Thread(new SocketListener(), "socket listener");
    listenthread.start();
  }

  public void shutdown()
  {
    running=false;
    listenthread.interrupt();
    Enumeration iterator=sockets.elements();
    try
    {
      while(iterator.hasMoreElements())
        ((Socket)iterator.nextElement()).close();
      listener.close();
    }
    catch(IOException e) {}
  }
}
