package momoko.server;

import java.net.*;
import java.util.*;
import momoko.*;
import java.io.IOException;
import momoko.drivers.*;
import momoko.stream.*;

public class StdioService implements Service, Runnable
{
  boolean running=true;
  Thread thread;

  String str;
  Class parserclass;
  Class formatterclass;
  Class interpreterclass;

    Parser parser;
    Formatter formatter;
    Interpreter interpreter;
    StreamHandler handler;

    public void run()
    {
      while(running)
      {
        try
        {
          parser=(Parser)parserclass.newInstance();
          formatter=(Formatter)formatterclass.newInstance();
          interpreter=(Interpreter)interpreterclass.newInstance();
          handler=new StreamHandler();
	  handler.setparser(parser);
	  handler.setformatter(formatter);
	  handler.setinstream(System.in);
	  handler.setoutstream(System.out);
	  handler.setinterpreter(interpreter);
	  handler.initialize();
          handler.handle();
        }
        catch(Exception e)
        {
            System.out.println("Error starting service on stdio");
            e.printStackTrace();
        }
      }
    }

  public void initialize(Properties args) throws
  UnspecifiedArguementException, BadArguementException, IOException,
  DriverNotFoundException
  {
    str=args.getProperty("parser");
    if(str==null)
      throw new UnspecifiedArguementException();
    try
    {
      parserclass=Class.forName(str);
    }
    catch(Exception e)
    {
        System.out.println("Could not load "+str);
        str="momoko.server."+str+"Parser";
        try {parserclass=Class.forName(str);}
        catch(Exception f) {throw new DriverNotFoundException(str);}
    }
    str=args.getProperty("interpreter");
    if(str==null)
      throw new UnspecifiedArguementException();
    try {interpreterclass=Class.forName("momoko.server."+str+"Interpreter");}
    catch(Exception e)
    {
      try
      {
        interpreterclass=Class.forName(str);
      }
      catch(Exception f) { throw new DriverNotFoundException(str); }
    }
    str=args.getProperty("formatter");
    if(str==null)
      throw new UnspecifiedArguementException();
    str="momoko.server."+str+"Formatter";
    try {formatterclass=Class.forName(str);}
    catch(Exception e) {throw new DriverNotFoundException(str);}
  }

  public void service()
  {
    thread=new Thread(this, "stdio listener");
    thread.start();
  }

  public void shutdown()
  {
    running=false;
    thread.interrupt();
  }
}
