package momoko.straw;

import momoko.util.*;
import java.io.*;
import momoko.compiler.Compiler;
import java.util.*;
import momoko.*;
import momoko.types.*;
import java.rmi.*;
import java.lang.reflect.*;
import java.net.*;

public class GenericStrawBroker extends MObject implements StrawBroker, LocalBroker, IMObject
{
  static StrawBroker instance=new GenericStrawBroker();
  static Hashtable references=new Hashtable();
  static Hashtable bindings=new Hashtable();
  static public boolean debug=false;

  public Serializable lookup(String path) throws RemoteException
  {
//if(debug) System.out.println("lookup("+path+")");
      Object obj=bindings.get(path);
//if(debug) System.out.println("lookup obj: "+obj);
      if(obj==null)
        throw new RemoteException("Not registered: "+path);
      else
        return export(obj);
    }

  public Serializable call(Integer i, String name, Object objs) throws RemoteException, NoSuchMethodException, IllegalAccessException, InvocationTargetException
  {
try
{
      if(debug) System.out.println("calling for "+name);
/*
if(name.equals("getName"))
{
  try { throw new NullPointerException(); }
  catch(Exception e) {e.printStackTrace();}
}
*/
      Object obj=references.get(i);
      if(obj==null)
      {
        if(debug) System.out.println("Bad reference!");
        if(debug) System.out.println(i);
        if(debug) System.out.println(references);
        throw new RemoteException("Bad reference: "+i.toString());
      }
      if(debug) System.out.println("the obj to call on is "+obj);
      if(debug) System.out.println("args: "+objs);

        Object[] objects=null;
        if(objs!=null)
        {
          Class cobj=objs.getClass();
          if(cobj.isArray())
          {
            objects=(Object[])objs;
            if(objects!=null)
              if(objects.length>0)
                objects=importArray(objects);
          }
          else if(objs instanceof StrawArrayReference)
          {
            objects=importArrayReference((StrawArrayReference)objs);
          }
          else
            throw new IllegalArgumentException();
        }

        Object o=null;
        try { o=invoke(obj, name, objects); }
        catch(Exception e) {e.printStackTrace();}
        if(debug) System.out.println("Success! "+name);
        if(o instanceof java.rmi.Remote)
        {
          if(debug) System.out.println("Object declared Remote, returning reference.");
          return export(o);
        }
        else
        {
          if(debug) System.out.println("Converting object to return it.");
          try { o=checkAndConvert(o); }
          catch(Exception e) {e.printStackTrace();}
          if(o==null)
          {
            if(debug)
              System.out.println("returning null");
          }
          else
          {
            if(debug)
              System.out.println("Returning "+o+" "+o.getClass()+" from "+name);
          }
          return (Serializable)o;
        }
}
catch(Exception e)
{
  e.printStackTrace();
  return null;
}
  }

  private Object invoke(Object obj, String name, Object[] objects) throws NoSuchMethodException, IllegalAccessException, InvocationTargetException
  {
    Class cls=obj.getClass();
      if(debug) System.out.println("calling "+cls.getName()+"."+name+"()");
      Method m=null;
      Class[] classes;
      if(objects==null)
        classes=null;
      else if(objects.length==0)
        classes=new Class[] {};
      else
      {
        classes=new Class[objects.length];
        for(int x=0;x<objects.length;x++)
        {
          if(objects[x] instanceof Proxy)
          {
//            if(debug) System.out.println("Proxy's primary class is 
// "+objects[x].getClass());
          }
          classes[x]=objects[x].getClass();
        }
/*
if(debug) System.out.println("interfaces(");
Class[] interfaces=cls.getInterfaces();
for(int x=0;x<interfaces.length;x++)
  if(debug) System.out.println(interfaces[x]);
if(debug) System.out.println(")");
if(debug) System.out.println("signature(");
for(int x=0;x<classes.length;x++)
  if(debug) System.out.println(classes[x]+" -- "+classes[x].getSuperclass());
if(debug) System.out.println(")");
if(debug) System.out.println("methods(");
Method[] ms=cls.getMethods();
for(int x=0;x<ms.length;x++)
  if(debug) System.out.println(ms[x]);
if(debug) System.out.println(")");
*/

      }

      try { m=cls.getMethod(name, classes); }
      catch(NoSuchMethodException e)
      { m=getMethodTheHardWay(cls, name, classes); }

//      if(debug) System.out.println(m);
      int mod=m.getModifiers();
//      if(debug) System.out.println(Modifier.toString(mod));

      if(obj instanceof Collection)
      {
        Collection c=(Collection)obj;
        return collectionKludge(c, m, name, objects);
      }

      if(obj instanceof Enumeration)
      {
        Enumeration e=(Enumeration)obj;
        return enumerationKludge(e, m, name, objects);
      }

      if(obj instanceof Iterator)
      {
        Iterator i=(Iterator)obj;
        return iteratorKludge(i, m, name, objects);
      }

      try { return m.invoke(obj, objects); }
      catch(Exception e) {return e;}
  }

  public Object collectionKludge(Collection obj, Method m, String name, Object[] objects) throws InvocationTargetException, IllegalAccessException
  {
    try
    {
    if(name.equals("iterator")) return obj.iterator();
    if(name.equals("toArray")) return obj.toArray();
    if(name.equals("toString")) return obj.toString();
    return m.invoke(obj, objects);
    }
    catch(IllegalAccessException e) {if(debug) System.out.println("Unkludged method: "+name); throw e; }
  }

  public Object iteratorKludge(Iterator obj, Method m, String name, Object[] objects) throws InvocationTargetException, IllegalAccessException
  {
    try
    {
    if(name.equals("hasNext")) return new Boolean(obj.hasNext());
    if(name.equals("next")) return obj.next();
    if(name.equals("toString")) return obj.toString();
    return m.invoke(obj, objects);
    }
    catch(IllegalAccessException e) {if(debug) System.out.println("Unkludged method: "+name); throw e;}
  }

  public Object enumerationKludge(Enumeration obj, Method m, String name, Object[] objects) throws InvocationTargetException, IllegalAccessException
  {
    try
    {
    if(name.equals("hasNext")) return new Boolean(((Iterator)obj).hasNext());
    if(name.equals("next")) return ((Iterator)obj).next();
    if(name.equals("hasMoreElements")) return new Boolean(obj.hasMoreElements());
    if(name.equals("nextElement")) return obj.nextElement();
    if(name.equals("toString")) return obj.toString();
    return m.invoke(obj, objects);
    }
    catch(IllegalAccessException e) {if(debug) System.out.println("Unkludged method: "+name); throw e;}
  }

  public Method getMethodTheHardWay(Class cls, String name, Class[] classes) throws IllegalAccessException, NoSuchMethodException
  {
    for(int x=0;x<classes.length;x++)
    {
      if(classes[x].equals(Integer.class))
        classes[x]=int.class;
      if(classes[x].equals(Boolean.class))
        classes[x]=boolean.class;
    }
    if(debug) System.out.println("Trying to find method "+name+" on "+cls.getName());
    if(debug) System.out.println(classes.length);
    if(debug) System.out.println(classes[0]);
    if(debug) System.out.println(classes[0].getName());
    if(debug) System.out.println(classes[0].getSuperclass());
    Class cz = classes[0].getSuperclass();
    if(cz!=null)
    {
      Class[] ints=cz.getInterfaces();
      for(int x=0;x<ints.length;x++)
        if(debug) System.out.println(ints[x]);
    }
    try { return cls.getMethod(name, classes); }
    catch(NoSuchMethodException e)
    {
      classes[0]=classes[0].getSuperclass().getInterfaces()[0];
      try { return cls.getMethod(name, classes); }
      catch(NoSuchMethodException f)
      {
        return getMethodExhaustively(cls, name, classes);
      }
    }
  }

  public Method getMethodExhaustively(Class cls, String name, Class[] classes)
  throws NoSuchMethodException
  {
    Method[] methods=cls.getMethods();
    Method m=null;
    for(int x=0;x<methods.length;x++) // For each possible matching method
    {
      m=methods[x];      
      if(!m.getName().equals(name)) continue;

      Class[] args=m.getParameterTypes();
      if(args.length!=classes.length) continue;

      for(int y=0;y<classes.length;y++) // For each arguement
      {
        boolean b=matchClass(classes[y], args[y]);
        if(debug) System.out.println(classes[y]+" ? "+args[y]+" = "+b+"("+y+"/"+classes.length+")");
        if(!b) { m=null; break; }
      }
      if(m!=null) return m;
    }
    throw new NoSuchMethodException(name);
  }

  public boolean matchClass(Class first, Class second)
  {
    if(debug) System.out.println(first+" ?= "+second);
    if(second.equals(Object.class)) return true;
    if(first.equals(second)) return true;
if(debug) System.out.println(second.isAssignableFrom(first));
if(debug) System.out.println(first.isAssignableFrom(second));
    if(first.isAssignableFrom(second)) return true;
    return false;    
  }

  public Serializable checkAndConvert(Object o)
  {
          if(check(o))
          {
//            if(debug) System.out.println("object is good as it is");
            Serializable s=(Serializable)o;
//            if(debug) System.out.println("checkAndConvert returning "+s);
            return s;
          }
          else
          {
//            if(debug) System.out.println("object must be converted");
            return convert(o);
          }
  }

  public Serializable convert(Object o)
  {
//          if(debug) System.out.println("Object "+o+" not serializable, sending reference.");
          Class c=o.getClass();
// System.out.println("c: "+c);
// System.out.println(c.isArray());
          if(c.isArray())
          {
            return convertArray((Object[])o);
          }
          else
            return export(o);
  }

  public Serializable convertArray(Object[] oarr)
  {
            if(oarr==null) return null;
            if(oarr.length==0) return (Serializable)oarr;
            Object o=oarr[0];
            if(check(o))
              return (Serializable)oarr;
            return exportArray(oarr);
  }

  public boolean check(Object o)
  {
    try
    {
//      File f=File.createTempFile("object", "ser");
//      OutputStream os=new FileOutputStream(f);
      OutputStream os=new BlackHoleStream();
      ObjectOutputStream oos=new ObjectOutputStream(os);
      oos.writeObject(o);
      oos.close();
//      f.delete();
      return true;
    }
    catch(NotSerializableException e)
    {
//      if(debug) System.out.println("I tried to serialize "+o+" but I 
// couldn't.");
      return false;
    }
    catch(Exception e)
    {
      e.printStackTrace();
      return false;
    }
  }

  public void bind(Object o, String s)
  {
    bindings.put(s, o);
  }

  public StrawReference export(Object o)
  {
//if(debug) System.out.println("Exporting "+o);
Class c=o.getClass();
Vector v=new Vector();
Class[] carr;
Hashtable h=new Hashtable();
while(c!=null)
{
//if(debug) System.out.println("class is "+c);
Class[] interfaces=c.getInterfaces();
for(int x=0;x<interfaces.length;x++)
{
//  if(debug) System.out.println(interfaces[x]);
  if(h.get(interfaces[x])==null)
  {
    v.addElement(interfaces[x]);
    h.put(interfaces[x], interfaces[x]);
  }
}
c=c.getSuperclass();
}
carr=new Class[v.size()];
for(int x=0;x<v.size();x++)
{
  carr[x]=(Class)v.elementAt(x);
}
          Integer i=new Integer(new Random().nextInt());
          references.put(i, o);
          return new StrawReference(i, o.getClass(), carr);
  }

  public StrawArrayReference exportArray(Object[] o)
  {
    //if(debug) System.out.println("Exporting "+o);
    Class c=o.getClass().getComponentType();
    Vector v=new Vector();
    Class[] carr;
    Hashtable h=new Hashtable();
    while(c!=null)
    {
      //if(debug) System.out.println("class is "+c);
      Class[] interfaces=c.getInterfaces();
      for(int x=0;x<interfaces.length;x++)
      { 
        //  if(debug) System.out.println(interfaces[x]);
        if(h.get(interfaces[x])==null)
        {
          v.addElement(interfaces[x]);
          h.put(interfaces[x], interfaces[x]);
        }
      }
      c=c.getSuperclass();
    }
    carr=new Class[v.size()];
    for(int x=0;x<v.size();x++)
    {
      carr[x]=(Class)v.elementAt(x);
    }
    StrawReference[] srarr=new StrawReference[o.length];
    for(int x=0;x<o.length;x++)
      srarr[x]=export(o[x]);
    return new StrawArrayReference(c, carr, srarr);
  }

  public Object[] importArray(Object[] objects)
  {
    Class c=objects.getClass().getComponentType();
    Object arr=Array.newInstance(c, objects.length);
    for(int x=0;x<objects.length;x++)
      Array.set(arr, x, importObject(objects[x]));
    return (Object[])arr;
  }

  public Object importObject(Object obj)
  {
    if(obj instanceof StrawReference)
    {
      StrawReference ref=(StrawReference)obj;
System.out.println("Importing object for reference: "+ref);
      boolean local=false;
      try { local=ref.isLocal(); } catch(Exception e) {e.printStackTrace();}
      if(debug) System.out.println("LOCAL: "+local);
      if(local)
        return references.get(ref.getId());
      else
      {
try{
        InetAddress addr=ref.getAddress();
        int port=ref.getPort();
System.out.println("Reference port: "+port);
        String str="//"+addr.getHostAddress()+":"+new Integer(port).toString()+"/StrawBroker";
System.out.println("Looking up broker at "+str);
        StrawBroker broker = (StrawBroker)java.rmi.Naming.lookup(str);
        return StrawProxy.newInstance(broker, ref);
}
catch(Exception e) {e.printStackTrace(); return null;}
      }
    }
    else if(obj instanceof StrawArrayReference)
      return importArrayReference((StrawArrayReference)obj);
    else
      return obj;
  }

  public Object[] importArrayReference(StrawArrayReference ref)
  {
System.out.println("ImportArrayReference");
    StrawReference[] refs=ref.getRefs();
    Object item=importObject(refs[0]);
    Object o=Array.newInstance(item.getClass(), refs.length);
    for(int x=0;x<refs.length;x++)
      Array.set(o, x, importObject(refs[x]));
    return (Object[])o;
  }
}
