package momoko.vfs;

import momoko.*;
import momoko.stream.*;
import momoko.protoclass.*;
import momoko.relationship.*;
import momoko.proxy.*;
import momoko.user.*;
import java.rmi.*;
import java.rmi.server.*;
import java.util.*;
import java.io.*;
import momoko.vfs.*;
import java.lang.reflect.*;
import java.beans.*;
import momoko.versioncontrol.*;

public class MObjectDirectory extends MObjectFile
{
  RealFile self;

public MObjectDirectory()
{
  this("");
}

public MObjectDirectory(String newname)
{
  super(newname);
  self=new RealFile(newname);
  if(!self.exists())
    self.mkdir();
}

public MObjectDirectory(File f, String newname)
{
  self=new RealFile(f, newname);
  if(!self.exists())
    self.mkdir();
}

public boolean release(IMObject object)  
{
  return true;
}

public boolean accept(IMObject object)  
{
  return object instanceof VirtualFile;
}

public void enterfunc(IMObject object)
{
  if(object instanceof File)
  {
    File f=(File)object;
    File dest=new File(self, object.getName());
    f.renameTo(dest);
  }
  else
  {
    System.out.println("Virtual files not yet handled.");
  }
}

public void exitfunc(IMObject object)
{
  File f=new File(self, object.getName());
  f.delete();
}

public VirtualFile getFileView()
{
  return null;
}

public VirtualFile getDirectoryView()
{
  return self;
}

public IMObject resolve(String path) throws BadPathException
{
if(debug) System.out.println("resolving: "+path);
  String prefix, suffix, temp=null;
  int i;
  boolean done=false;
  Iterator iterator;
 
  i=path.indexOf('/');
  if(i==-1)
  {
    prefix=path;
    suffix="";
  }
  else if(i==0)
  {
    return Database.main.root.resolve(path.substring(1));
  }
  else
  {
    prefix = path.substring(0, i);
    suffix = path.substring(i+1);
  }
  File next=new File(self, prefix);

  if(!next.exists()) throw new BadPathException(this.getName()+"/"+path);

  IMObject obj;
  if(next.isDirectory()) obj=new MObjectDirectory(self, prefix);
  else obj=new MObjectFile(self, prefix);

  if(suffix.equals(""))
    return obj;
  else
    return obj.resolve(suffix);
}

public void remove(String str)
{
  File f=new File(self, str);
  f.delete();
}

public Map contents()
{
  Map map=new HashMap();
  File[] files=self.listFiles();
  for(int x=0;x<files.length;x++)
  {
    File f=files[x];
    if(f.isDirectory())
      map.put(f.getName(), new MObjectDirectory(self, f.getName()));
    else
      map.put(f.getName(), new MObjectFile(self, f.getName()));
  }
  return map;
}

public void setcontents(Map newcontents)
{
  throw new NullPointerException();
}

public void put(IMObject obj)
{
  if(obj instanceof File)
  {
    File f=(File)obj;
    File dest=new File(self, obj.getName());
    f.renameTo(dest);
  }
}

}
